from .positional_encoding import PositionalEncoding
from .mlp import MLP
from .mlp import PositionWiseFeedForward
from .se_layer import SELayer1d
from .se_layer import SELayer2d
from .dense_interpolation import DenseInterpolation

__all__ = [
    "PositionalEncoding", "MLP", "PositionWiseFeedForward", "SELayer1d", "SELayer2d", "DenseInterpolation"
]



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import io
import time
from pathlib import Path
from copy import deepcopy
from abc import ABC, abstractmethod
from collections import OrderedDict

import numpy as np
import torch
import torch.nn as nn
from sklearn import base
from torch.utils.data import DataLoader
from torch.utils.data import SubsetRandomSampler

from enchanter.engine import modules

if modules.is_jupyter():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


__all__ = [
    "BaseRunner"
]


class BaseRunner(base.BaseEstimator, ABC):
    """
    PyTorchモデルの訓練に用いる Runner を作成する為のクラスです。


    Examples:

        >>> class Runner(BaseRunner):
        >>>     def __init__(self):
        >>>         super(Runner, self).__init__()
        >>>         self.model = nn.Linear(10, 10)
        >>>         self.optimizer = torch.optim.Adam(self.model.parameters())
        >>>         self.experiment = Experiment()
        >>>         self.criterion = nn.CrossEntropyLoss()
        >>>
        >>>     def train_step(self, batch):
        >>>         x, y = batch
        >>>         out = self.model(x)
        >>>         loss = self.criterion(out, y)
        >>>
        >>>         return {"loss": loss}

    """
    def __init__(self):
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = NotImplemented
        self.optimizer = NotImplemented
        self.scheduler = None
        self.experiment = NotImplemented
        self.early_stop = None

        self._epochs = 1
        self.pbar = None
        self._loaders = {}
        self._metrics = {}
        self._checkpoint_path = None

    @abstractmethod
    def train_step(self, batch):
        """
        ニューラルネットの訓練時、
            >>> for x, y in train_loader:
            >>>     out = model(x)
            >>>     loss = criteion(out, y)

        にあたる箇所を担当するメソッドです。

        Args:
            batch: PyTorch DataLoader から得られる訓練用のデータトラベルを含むタプル

        Returns:
            キーに 'loss' を含む辞書を返す必要があります。

        Examples:
            >>> def train_step(self, batch):
            >>>     x, y = batch
            >>>     out = self.model(x)
            >>>     loss = nn.functional.cross_entropy(out, y)
            >>>     return {"loss": loss}

        """

    def train_end(self, outputs):
        """
        ニューラルネットの訓練時、1step 終了ごとに実行されるメソッドです。

        Args:
            outputs:

        Returns:

        """
        return {}

    def val_step(self, batch):
        """
        ニューラルネットの検証時、1step ごとに実行されるメソッドです。

        Args:
            batch: PyTorch DataLoader から得られる訓練用のデータトラベルを含むタプル

        Returns:
            辞書を返す必要があります。

        """

    def val_end(self, outputs):
        """
        ニューラルネットの検証時、1step 終了ごとに実行されるメソッドです。

        Args:
            outputs:

        Returns:

        """
        return {}

    def test_step(self, batch):
        """
        ニューラルネットの評価時、1step ごとに実行されるメソッドです。

        Args:
            batch: PyTorch DataLoader から得られる訓練用のデータトラベルを含むタプル

        Returns:
            辞書を返す必要があります。

        """

    def test_end(self, outputs):
        """
        ニューラルネットの評価時、1step 終了ごとに実行されるメソッドです。

        Args:
            outputs:

        Returns:

        """
        return {}

    def train_cycle(self, epoch, loader):
        """
        ニューラルネットの訓練ループです。

        Args:
            epoch (int):
            loader (torch.utils.data.DataLoader):

        """
        results = list()
        loader_size = len(loader)

        self.model.train()
        with self.experiment.train():
            for step, batch in enumerate(loader):
                self.optimizer.zero_grad()
                batch = tuple(map(lambda x: x.to(self.device), batch))
                # on_step_start()
                outputs = self.train_step(batch)
                outputs["loss"].backward()
                self.optimizer.step()

                per = "{:1.0%}".format(step / loader_size)
                self.pbar.set_postfix(
                    OrderedDict(train_batch=per), refresh=True
                )

                outputs = {
                    key: outputs[key].detach().cpu() if isinstance(outputs[key], torch.Tensor) else outputs[key]
                    for key in outputs.keys()
                }
                self.experiment.log_metrics(outputs, step=step, epoch=epoch)
                results.append(outputs)
                # on_step_end()

            dic = self.train_end(results)        # pylint: disable=E1111

            if len(dic) != 0:
                self._metrics.update(dic)
                self.experiment.log_metrics(dic, step=epoch)

    def val_cycle(self, epoch, loader):
        """
        ニューラルネットの評価用ループです。

        Args:
            epoch:
            loader:

        Returns:

        """
        results = list()
        loader_size = len(loader)

        self.model.eval()
        with self.experiment.validate():
            with torch.no_grad():
                for step, batch in enumerate(loader):
                    batch = tuple(map(lambda x: x.to(self.device), batch))
                    # on_step_start()
                    outputs = self.val_step(batch)        # pylint: disable=E1111

                    per = "{:1.0%}".format(step / loader_size)
                    self.pbar.set_postfix(
                        OrderedDict(val_batch=per), refresh=True
                    )

                    outputs = {
                        key: outputs[key].cpu() if isinstance(outputs[key], torch.Tensor) else outputs[key]
                        for key in outputs.keys()
                    }
                    self.experiment.log_metrics(outputs, step=step, epoch=epoch)
                    results.append(outputs)
                    # on_step_end()

                dic = self.val_end(results)        # pylint: disable=E1111

                if len(dic) != 0:
                    self._metrics.update(dic)
                    self.experiment.log_metrics(dic, step=epoch)

    def test_cycle(self, loader):
        """
        ニューラルネットの検証用ループです。

        Args:
            loader:

        Returns:

        """
        results = list()
        loader_size = len(loader)

        self.model.eval()
        with self.experiment.test():
            with torch.no_grad():
                for step, batch in enumerate(loader):
                    batch = tuple(map(lambda x: x.to(self.device), batch))
                    # on_step_start()
                    outputs = self.test_step(batch)        # pylint: disable=E1111

                    per = "{:1.0%}".format(step / loader_size)
                    self.pbar.set_postfix(
                        OrderedDict(test_batch=per), refresh=True
                    )

                    outputs = {
                        key: outputs[key].cpu() if isinstance(outputs[key], torch.Tensor) else outputs[key]
                        for key in outputs.keys()
                    }

                    self.experiment.log_metrics(outputs, step=step)
                    results.append(outputs)
                    # on_step_end()

                dic = self.test_end(results)        # pylint: disable=E1111

                if len(dic) != 0:
                    self._metrics.update(dic)
                    self.experiment.log_metrics(dic)

    def train_config(self, epochs, *args, **kwargs):
        """
        .run() メソッドを用いて訓練を行う際に、 epochs などを指定する為のメソッドです。

        Args:
            epochs (int):
            *args:
            **kwargs:

        Returns:

        """

        self._checkpoint_path = kwargs.get("checkpoint_path", None)

        if epochs > 0:
            self._epochs = epochs
        else:
            self._epochs = 1

        return self

    def log_hyperparams(self, dic=None, prefix=None):
        """

        Args:
            dic (Dict):
            prefix (str):

        Returns:
            None
        """
        self.experiment.log_parameters(self.optimizer.__dict__["defaults"], prefix="optimizer")
        self.experiment.log_parameter("Optimizer", self.optimizer.__class__.__name__)

        if dic is not None:
            self.experiment.log_parameters(dic, prefix)

    def initialize(self):
        """
        Runner の準備を行うメソッドです。
        self.model, self.optimizer, self.experiment といった実行に必要な最低限の変数が未定義な場合はエラーメッセージを出し終了します。
        問題がなければ CPU or GPU にモデルを渡します。

        Returns:
            None
        """
        if self.model is None:
            raise Exception("self.model is not defined.")

        if self.optimizer is None:
            raise Exception("self.optimizer is not defined.")

        if self.experiment is None:
            raise Exception("self.experiment is not defined.")

        self.model = self.model.to(self.device)

    def run(self, verbose=True):
        """
        Runnerを実行します。
        実行には、事前に self.add_loader("train", train_loader) で訓練用のデータローダを登録するしておく必要があります。

        Args:
            verbose: True の場合、学習の進行を表示します。

        Returns:
            None
        """
        self.initialize()
        self.log_hyperparams()

        if not self.loaders:
            raise Exception("At least one DataLoader must be provided.")

        if "train" in self.loaders:
            self.pbar = tqdm(range(self._epochs), desc="Epochs") if verbose else range(self._epochs)
            # .on_epoch_start()
            for epoch in self.pbar:
                # on_train_start()
                self.train_cycle(epoch, self.loaders["train"])
                # on_train_end()

                if "val" in self.loaders:
                    # on_validation_start()
                    self.val_cycle(epoch, self.loaders["val"])
                    # on_validation_end()

                if self.scheduler:
                    self.scheduler.step(epoch=None)
                    self.experiment.log_metric("scheduler_lr", self.scheduler.get_last_lr(), epoch=epoch)

                if self.early_stop:
                    if self.early_stop.on_epoch_end(self._metrics, epoch):
                        break
                    # .on_epoch_end()

                if self._checkpoint_path:
                    self.save(self._checkpoint_path, epoch=epoch)

        if "test" in self.loaders:
            # on_test_start()
            self.test_cycle(self.loaders["test"])
            # on_test_end()

        return self

    def predict(self, x):
        """
        与えられた入力をもとに予測を行うメソッドです。

        Args:
            x (Union[torch.Tensor, np.ndarray]):

        Returns:
            predict
        """
        raise NotImplementedError

    def add_loader(self, mode, loader):
        """
        訓練等に用いるデータローダをRunnerに登録する為のメソッドです。

        Args:
            mode (str): ['train', 'val', 'test'] のいずれをか指定します。
            loader (torch.utils.data.DataLoader):

        """
        if mode not in ["train", "val", "test"]:
            raise Exception("argument `mode` must be one of 'train', 'val', or 'test'.")

        if not isinstance(loader, DataLoader):
            raise Exception("The argument `loader` must be an instance of `torch.utils.data.DataLoader`.")

        self.experiment.log_parameters(loader.__dict__, prefix=mode)
        self.experiment.log_parameter("{}_dataset_len".format(mode), len(loader))

        self._loaders[mode] = loader
        return self

    @property
    def loaders(self):
        return self._loaders

    def fit(self, x, y, **kwargs):
        """
        Scikit-Learn スタイルの訓練メソッドです。

        Args:
            x: 訓練用データ
            y: 教師ラベル
            **kwargs:

        """
        val_size: float = kwargs.get("val_size", 0.1)
        num_workers = kwargs.get("num_workers", 0)
        batch_size = kwargs.get("batch_size", 1)
        pin_memory = kwargs.get("pin_memory", False)
        verbose = kwargs.get("verbose", True)

        if self._epochs == 0:
            epochs = kwargs.get("epochs", 1)
        else:
            epochs = self._epochs

        train_ds = modules.get_dataset(x, y)
        val_ds = modules.get_dataset(x, y)
        n_train = len(train_ds)
        indices = list(range(n_train))
        split = int(np.floor(val_size * n_train))

        train_idx, val_idx = indices[split:], indices[:split]
        train_sampler = SubsetRandomSampler(train_idx)
        val_sampler = SubsetRandomSampler(val_idx)

        train_loader = DataLoader(
            train_ds, batch_size=batch_size,
            sampler=train_sampler, num_workers=num_workers,
            pin_memory=pin_memory
        )
        val_loader = DataLoader(
            val_ds, batch_size=batch_size,
            sampler=val_sampler, num_workers=num_workers, pin_memory=pin_memory
        )

        self.add_loader("train", train_loader)
        self.add_loader("val", val_loader)
        self.train_config(epochs, checkpoint_path=self._checkpoint_path)
        self.run(verbose)

        return self

    def freeze(self):
        """
        モデルのパラメータが勾配を計算しないように固定する為のメソッドです。

        """
        for param in self.model.parameters():
            param.requires_grad = False

        self.model.eval()

    def unfreeze(self):
        """
        .freeze() で固定されたパラメータを再度学習できるようにする為のメソッドです。

        """
        for param in self.model.parameters():
            param.requires_grad = True

        self.model.train()

    def save_checkpoint(self):
        """
        ニューラルネットの重みと、 Optimizerの状態を辞書として出力するメソッドです。

        Returns:
            以下のキーと値を持つ辞書を返します。
                - "model_state_dict": ニューラルネットの重み
                - "optimizer_state_dict": Optimizer の状態

        """
        if isinstance(self.model, nn.DataParallel):
            model = self.model.module.state_dict()
        else:
            model = self.model.state_dict()

        checkpoint = {
            "model_state_dict": deepcopy(model),
            "optimizer_state_dict": deepcopy(self.optimizer.state_dict())
        }
        return checkpoint

    def load_checkpoint(self, checkpoint):
        """
        'model_state_dict', 'optimizer_state_dict' を持つ辞書を受け取り、それらを元に モデル と Optimizer の状態を復元します。

        Args:
            checkpoint:
                以下のキーと値を持つ辞書。
                    - "model_state_dict": ニューラルネットの重み
                    - "optimizer_state_dict": Optimizer の状態
        """
        self.model.load_state_dict(checkpoint["model_state_dict"])
        self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        return self

    def save(self, directory=None, epoch=None):

        """
        指定したディレクトリにモデルとOptimizerの状態を記録したファイルを保存します。

        Args:
            directory (Optional[str]):
            epoch (Optional[int]):

        """
        if directory is None and self._checkpoint_path is not None:
            directory = self._checkpoint_path

        elif directory is None and self._checkpoint_path is None:
            raise ValueError("The argument `directory` must be specified.")

        directory = Path(directory)
        if not directory.exists():
            directory.mkdir(parents=True)
        checkpoint = self.save_checkpoint()

        if epoch is None:
            epoch = time.ctime().replace(" ", "_")

        filename = "checkpoint_epoch_{}.pth".format(epoch)
        path = directory / filename
        torch.save(checkpoint, path)

        if hasattr(self.experiment, "log_asset_data"):
            buffer = io.BytesIO()
            torch.save(checkpoint, buffer)
            self.experiment.log_asset_data(buffer.getvalue(), filename)

    def load(self, filename, map_location="cpu"):
        """
        指定したファイルを元にモデルとOptimizerの状態を復元します。

        Args:
            filename (str):
            map_location (str):

        """
        checkpoint = torch.load(filename, map_location=map_location)
        self.load_checkpoint(checkpoint)

        return self



from .classification import ClassificationRunner
from .regression import RegressionRunner
from .ensemble import HardEnsemble, SoftEnsemble

__all__ = [
    "ClassificationRunner", "RegressionRunner",
    "HardEnsemble", "SoftEnsemble"
]



from sklearn.metrics import roc_curve
from scipy.optimize import brentq
from scipy.interpolate import interp1d


__all__ = [
    "calculate_eer"
]


def calculate_eer(y_true, y_score):
    """
    Returns the equal error rate for a binary classification output.
    The equal error rate is the value of FPR (or FNR) when the ROC curves intersects the line connecting (0,0) to (1,1).

    References:
        `Should sklearn include the Equal Error Rate metric \
            <https://github.com/scikit-learn/scikit-learn/issues/15247>`_

    Args:
        y_true (np.ndarray (1D, float)):
        y_score (np.ndarray (1D, float)):

    Returns:
        return EER Score
    """
    fpr, tpr, thresholds = roc_curve(y_true, y_score, pos_label=1)
    eer_score = brentq(lambda x: 1. - x - interp1d(fpr, tpr)(x), 0., 1.)
    return eer_score



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import torch
import torch.nn as nn


__all__ = [
    "MLP", "PositionWiseFeedForward"
]


class MLP(nn.Module):
    """
    MLPを作成するクラス

    Examples:
        >>> import enchanter.addons as addons
        >>> model = addons.layers.MLP([10, 512, 128, 5], addons.Mish())
        >>> print(model)
        >>> # ModuleList(
        >>> #    (0): Linear(in_features=10, out_features=512, bias=True)
        >>> #    (1): Linear(in_features=512, out_features=128, bias=True)
        >>> #    (2): Linear(in_features=128, out_features=5, bias=True)
        >>> #)
    """
    def __init__(self, shapes, activation=torch.relu):
        """

        Args:
            shapes (List[int]): MLPの各層におけるニューロン数。int型の要素で構成される配列を想定。
                    与えられる配列の第0番目の要素の値はモデルへの入力次元の数として扱われます。

            activation (Union[Callable[[torch.Tensor], torch.Tensor], nn.Module]): 活性化関数。
                        torch.relu や enchanter.addons.Mish() 等の微分可能な Callableなオブジェクト
        """
        super().__init__()
        self.layers = []
        self.activation = activation

        for i in range(len(shapes) - 1):
            self.layers.append(nn.Linear(shapes[i], shapes[i+1]))

        self.layers = nn.ModuleList(self.layers)

    def forward(self, x):
        for layer in self.layers[:-1]:
            x = layer(x)
            x = self.activation(x)

        x = self.layers[-1](x)
        return x


class PositionWiseFeedForward(nn.Module):
    """
    Attention is all you need. で提案された PositionWiseFeedForward の 1×1 Conv1d を使ったバージョン

    Examples:
        >>> import torch
        >>> import enchanter.addons as addons
        >>> x = torch.randn(1, 128, 512)    # [N, seq_len, features]
        >>> ff = addons.layers.PositionWiseFeedForward(512)
        >>> out = ff(x)
    """
    def __init__(self, hidden_size):
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv1d(hidden_size, hidden_size*2, 1),
            nn.ReLU(),
            nn.Conv1d(hidden_size*2, hidden_size, 1)
        )

    def forward(self, x):
        """
        入力に対して PositionWiseFeedForward を適用します。

        Args:
            x (torch.Tensor):

        Returns:

        """
        x = x.transpose(1, 2)
        x = self.conv(x)
        x = x.transpose(1, 2)
        return x



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from contextlib import contextmanager
import torch
import numpy as np
from torch.utils.tensorboard import SummaryWriter


__all__ = [
    "BaseLogger", "TensorBoardLogger"
]


class BaseLogger:
    def __init__(self):
        self.context = None

    @contextmanager
    def train(self):
        old_state = self.context
        self.context = "train"

        yield self

        self.context = old_state

    @contextmanager
    def validate(self):
        old_state = self.context
        self.context = "validate"

        yield self

        self.context = old_state

    @contextmanager
    def test(self):
        old_state = self.context
        self.context = "test"

        yield self

        self.context = old_state

    def log_metric(self, name, value, step=None, epoch=None, include_context=True):
        pass

    def log_metrics(self, dic, prefix=None, step=None, epoch=None):
        pass

    def log_parameter(self, name, value, step=None):
        pass

    def log_parameters(self, dic, prefix=None, step=None):
        pass

    def set_model_graph(self, *args, **kwargs):
        pass


class TensorBoardLogger(BaseLogger):
    def __init__(self, log_dir, *args, **kwargs):
        super().__init__()
        self.writer = SummaryWriter(log_dir, *args, **kwargs)

    def log_metric(self, name, value, step=None, epoch=None, include_context=True):
        self.writer.add_scalar("{}/{}".format(self.context, name), value)

    def log_metrics(self, dic, prefix=None, step=None, epoch=None):
        for k, v in dic.items():
            self.writer.add_scalar(k, v)

    def log_parameter(self, name, value, step=None):
        if isinstance(value, torch.Tensor) or\
                isinstance(value, np.ndarray) or\
                isinstance(value, float) or\
                isinstance(value, int):
            self.writer.add_scalar("{}/{}/{}".format(self.context, "params", name), value, step)

    def log_parameters(self, dic: dict, prefix=None, step=None):
        for k, v in dic.items():
            self.log_parameter(k, v, step)



# *******************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# *******************************************************

from typing import List

import torch
import numpy as np
from sklearn.base import BaseEstimator

from enchanter.engine.runner import BaseRunner
from enchanter.engine import modules


__all__ = [
    "BaseEnsembleEstimator", "SoftEnsemble", "HardEnsemble"
]


class BaseEnsembleEstimator(BaseEstimator):
    def __init__(self, runners: List[BaseRunner], mode: str = None):
        self.runners = runners
        self.mode = mode
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    def predict(self, x) -> List[np.ndarray]:
        x = modules.numpy2tensor(x).to(self.device)

        predicts = []
        for runner in self.runners:
            predicts.append(runner.predict(x))

        return predicts


class SoftEnsemble(BaseEnsembleEstimator):
    """
    確率の平均をとるアンサンブル
    """

    def predict(self, x) -> np.ndarray:
        """

        Args:
            x:

        Returns:

        """
        predicts = super().predict(x)

        predicts = sum(predicts)
        probs = predicts / len(self.runners)

        if self.mode == "classification":
            return probs.astype(np.int)
        else:
            return probs


class HardEnsemble(BaseEnsembleEstimator):
    """
    多数決をとるアンサンブル
    """
    def __init__(self, runners: List):
        super().__init__(runners, mode="classification")

    def predict(self, x) -> np.ndarray:
        """

        Args:
            x:

        Returns:

        """
        predicts = super().predict(x)
        predicts = np.stack(predicts)

        return np.ravel(predicts[0])



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

from typing import Dict
import numpy as np
from .base import Callback


__all__ = [
    "EarlyStopping"
]


class EarlyStopping(Callback):
    def __init__(self, monitor: str = "validate_avg_loss", min_delta=0.0, patience: int = 0, mode: str = "auto"):
        Callback.__init__(self)
        self.monitor = monitor
        self.patience = patience
        self.min_delta = min_delta
        self.wait = 0
        self.stopped_epoch = 0

        mode_dict = {
            "min": np.less,
            "max": np.greater,
            "auto": np.greater if "acc" in self.monitor else np.less
        }
        if mode not in mode_dict:
            mode = "auto"

        self.monitor_op = mode_dict[mode]
        self.min_delta *= 1 if self.monitor_op == np.greater else -1
        self.best = np.Inf if self.monitor_op == np.less else -np.Inf

    def check_metrics(self, logs):
        monitor_val = logs.get(self.monitor)

        if monitor_val is None:
            return False

        return True

    def on_epoch_end(self, metrics: Dict, epoch: int) -> bool:
        stop = False
        if not self.check_metrics(metrics):
            return stop

        current = metrics.get(self.monitor)
        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                stop = True

        return stop



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************

import random
import torch
import numpy as np
from torch.utils.data import TensorDataset


__all__ = [
    "is_jupyter", "numpy2tensor", "get_dataset", "fix_seed"
]


def is_jupyter():
    """
    実行中の環境が Jupyter Notebookかどうかを判定します。

    Returns (bool) :
        True if run on jupyrer notebook else False

    """
    if "get_ipython" not in globals():
        return False
    env = get_ipython().__class__.__name__      # NOQA
    if env == "TerminalInteractiveShell":
        return False
    return True


def numpy2tensor(inputs):
    """
    入力された `np.ndarray` を `torch.Tensor` に変換します。単に `torch.Tensor` が入力された場合は、そのまま値を返します。

    Examples:
        >>> x = np.random.randn(512, 6)     # np.ndarray (dtype=np.float64)
        >>> x = numpy2tensor(x)             # torch.Tensor (dtype=torch.DoubleTensor)

    Args:
        inputs (Union[np.ndarray, torch.Tensor]):

    Returns:
        `torch.Tensor`
    """
    if isinstance(inputs, np.ndarray):
        inputs = torch.from_numpy(inputs)

    return inputs


def get_dataset(x, y=None):
    """
    入力された値をもとに `torch.utils.data.TensorDataset` を生成します。

    Examples:
        >>> x = torch.randn(512, 6)
        >>> y = torch.randint(0, 9, size=[512])
        >>> ds = get_dataset(x, y)

    Args:
        x (Union[np.ndarray, torch.Tensor]):
        y (Optional[Union[np.ndarray, torch.Tensor]]):

    Returns:
        `torch.utils.data.TensorDataset`
    """
    x = numpy2tensor(x)

    if y is not None:
        y = numpy2tensor(y)
        ds = TensorDataset(x, y)
    else:
        ds = TensorDataset(x)

    return ds


def fix_seed(seed):
    """
    PyTorch, NumPy, Pure Python Random のSEED値を一括固定します。

    Examples:
        >>> fix_seed(0)
        >>> x = torch.randn(...)
        >>> y = np.random.randn(...)

    Args:
        seed (int): SEED値

    Returns:
        None
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)



import os
from setuptools import setup

PATH_ROOT = os.path.dirname(__file__)


def load_requirements(path_dir=PATH_ROOT, comment_char='#'):
    with open(os.path.join(path_dir, 'requirements.txt'), 'r') as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for ln in lines:
        # filer all comments
        if comment_char in ln:
            ln = ln[:ln.index(comment_char)]
        if ln:  # if requirement is not empty
            reqs.append(ln)
    return reqs


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='enchanter',
    version='0.5.0',
    packages=[
        'enchanter', 'enchanter.addons', 'enchanter.addons.layers',
        'enchanter.callbacks', 'enchanter.engine', 'enchanter.metrics',
        'enchanter.preprocessing', 'enchanter.wrappers', 'enchanter.utils'
    ],
    url='https://github.com/khirotaka/enchanter',
    license='Apache-2.0',
    author='Hirotaka Kawashima',
    author_email='',
    description='Enchanter is a library for machine learning tasks for comet.ml users.',
    install_requires=load_requirements(),
    long_description=long_description,
    long_description_content_type='text/markdown'
)



# ***************************************************
#  _____            _                 _
# | ____|_ __   ___| |__   __ _ _ __ | |_ ___ _ __
# |  _| | '_ \ / __| '_ \ / _` | '_ \| __/ _ \ '__|
# | |___| | | | (__| | | | (_| | | | | ||  __/ |
# |_____|_| |_|\___|_| |_|\__,_|_| |_|\__\___|_|
#
# ***************************************************


from pprint import pformat
import pkg_resources


if "comet-ml" not in [d.project_name for d in pkg_resources.working_set]:
    raise ImportError("You have to install `comet_ml` if you want to use `enchanter.utils.comet` module.")


__all__ = [
    "TunerConfigGenerator"
]


class TunerConfigGenerator:
    def __init__(
            self, algorithm="bayes", metric="validate_avg_loss", objective="minimize", seed=None, max_combo=0,
            grid_size=10, min_sample_size=100, retry_limit=20,  retry_assign_limit=0, name=None, trials=1
    ):
        """
        各引数のより詳しい説明は https://www.comet.ml/docs/python-sdk/introduction-optimizer/ を参照してください。
        
        Args:
            algorithm: パラメータチューニングに用いるアルゴリズムを指定します。対応しているアルゴリズムは ['grid', 'random', 'bayes'] です。
            metric: 最小化/最大化する値を指定します。デフォルトでは、'validate_avg_loss' が指定されています。
            objective: metrics を最大化/最小化するかを指定します。['minimize', 'maximize'] で指定してください。
            seed: シード値を設定します。デフォルトでは指定されていません。
            max_combo: 整数、試行するパラメーターの組み合わせの制限（デフォルトは0、制限なしを意味します）
            grid_size: 整数、グリッドを作成するときのパラメーターあたりのビンの数（デフォルトは10）
            min_sample_size: 整数、適切なグリッド範囲を見つけるのに役立つサンプル数（デフォルトは100）
            retry_limit: 整数、中断する前に一意のパラメータセットを作成しようとする制限（デフォルトは20）
            retry_assign_limit:
            name: 文字列、この検索インスタンスに関連付けるパーソナライズ可能な名前（オプション）
            trials: 試行回数を指定します。
        """

        if algorithm not in ["grid", "bayes", "random"]:
            raise ValueError("The algorithms you can select are `random`, `bayes` and `grid`.")

        self.algorithm = algorithm
        self.spec = {
            "maxCombo": max_combo,
            "objective": objective,
            "metric": metric,
            "seed": seed,
            "gridSize": grid_size,
            "minSampleSize": min_sample_size,
            "retryLimit": retry_limit,
            "retryAssignLimit": retry_assign_limit
        }
        self.name = name
        self.trials = trials
        self.__params = {}

    def suggest_categorical(self, name, values):
        """
        カテゴリカル変数を探索する為のメソッドです。

        Args:
            name: 変数の名前を指定します。
            values: 探索する変数をしていします。comet.mlの仕様上、文字列のリストを与える必要があります。

        Examples:
            >>> import comet_ml
            >>> from enchanter.utils.comet import TunerConfigGenerator
            >>> config = TunerConfigGenerator()
            >>> config.suggest_categorical("activation", ["torch.relu", "torch.sigmoid", "torch.softmax"])
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     activation = experiment.get_parameter(eval("activation"))

        """
        self.__params["{}".format(name)] = {
            "type": "categorical",
            "values": [str(value) for value in values]
        }
        return self

    def __suggest(self, name, min_value, max_value, dtype, scaling, **kwargs):
        if dtype is None:
            if type(min_value) is type(max_value):
                dtype = type(min_value)
            else:
                raise Exception("`min_value` and `max_value` must be same data type.")

        if dtype is float:
            dtype = "float"
        elif dtype is int:
            dtype = "integer"
        else:
            raise Exception("`dtype` must be specified as `int` or `float` .")

        self.__params["{}".format(name)] = {
            "type": dtype,
            "min": min_value,
            "max": max_value,
            "scalingType": scaling
        }

        for key in ["mu", "sigma"]:
            if key in kwargs:
                self.__params["{}".format(name)][key] = kwargs[key]

    def suggest_linear(self, name, min_value, max_value, dtype=None):
        """

        Args:
            name:
            min_value:
            max_value:
            dtype:

        Returns:

        """
        self.__suggest(name, min_value, max_value, dtype, "linear")
        return self

    def suggest_uniform(self, name, min_value, max_value, dtype=None):
        """
        変数を一様分布からサンプリングし探索する為のメソッドです。

        Args:
            name: 変数の名前
            min_value: 最小値
            max_value: 最大値
            dtype: データ型。指定されない場合、`min_value` と `max_value` の値から自動的に推定されます。

        Examples:
            >>> config = TunerConfigGenerator()
            >>> config.suggest_uniform("uniform", 0.0, 1.0)
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     uniform = experiment.get_parameter("uniform")

        """
        self.__suggest(name, min_value, max_value, dtype, "uniform")
        return self

    def suggest_normal(self, name, min_value, max_value, mu=0.0, sigma=1.0, dtype=None):
        """
        変数を正規分布からサンプリングし探索する為のメソッドです。

        Args:
            name: 変数の名前
            min_value: 最小値
            max_value: 最大値
            mu: 正規分布のμ
            sigma: 正規分布のσ
            dtype: データ型。指定されない場合、`min_value` と `max_value` の値から自動的に推定されます。

        Examples:
            >>> config = TunerConfigGenerator()
            >>> config.suggest_normal("normal", 0.0, 1.0)
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     normal = experiment.get_parameter("normal")

        """
        self.__suggest(name, min_value, max_value, dtype, "normal", mu=mu, sigma=sigma)
        return self

    def suggest_lognormal(self, name, min_value, max_value, mu=0.0, sigma=1.0, dtype=None):
        """
        変数を対数正規分布からサンプリングし探索する為のメソッドです。

        Args:
            name: 変数の名前
            min_value: 最小値
            max_value: 最大値
            mu: 正規分布のμ
            sigma: 正規分布のσ
            dtype: データ型。指定されない場合、`min_value` と `max_value` の値から自動的に推定されます。

        Examples:
            >>> config = TunerConfigGenerator()
            >>> config.suggest_lognormal("lognormal", 0.0, 1.0)
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     lognormal = experiment.get_parameter("lognormal")

        """
        self.__suggest(name, min_value, max_value, dtype, "lognormal", mu=mu, sigma=sigma)
        return self

    def suggest_loguniform(self, name, min_value, max_value, dtype=None):
        """
        変数を対数一様分布からサンプリングし探索する為のメソッドです。

        Args:
            name: 変数の名前
            min_value: 最小値
            max_value: 最大値
            dtype: データ型。指定されない場合、`min_value` と `max_value` の値から自動的に推定されます。

        Examples:
            >>> config = TunerConfigGenerator()
            >>> config.suggest_loguniform("loguniform", 0.0, 1.0)
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     loguniform = experiment.get_parameter("loguniform")

        """
        self.__suggest(name, min_value, max_value, dtype, "loguniform")
        return self

    def suggest_discrete(self, name, values):
        """
        指定した数値型の変数を探索する為のメソッドです。

        Args:
            name: 変数の名前
            values: 数値型の要素で構成されたリスト

        Examples:
            >>> config = TunerConfigGenerator()
            >>> config.suggest_discrete("discrete", [10, 20, 30])
            >>> opt = comet_ml.Optimizer(config.generate())
            >>> for experiment in opt.get_experiments():
            >>>     discrete = experiment.get_parameter("discrete")

        """
        self.__params["{}".format(name)] = {
            "type": "discrete",
            "values": values
        }
        return self

    def generate(self):
        """
        `comet_ml.Optimizer` 用のConfigを生成する為のメソッドです。

        Examples:
            >>> config = TunerConfigGenerator()
            >>> config.suggest_discrete("discrete", [10, 20, 30])
            >>> cfg = config.generate()
        """

        config = {
            "algorithm": self.algorithm,
            "parameters": self.__params,
            "spec": self.spec,
            "trials": self.trials,
            "name": self.name
        }
        return config

    def export(self, filename):
        """
        作成したConfigをファイルにして保存する為のメソッドです。

        """
        with open(filename, "w") as f:
            f.write(self.__repr__())

    def __repr__(self):
        return pformat(self.generate(), indent=1)
